
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <dirent.h>
#include <signal.h>
#include <netdb.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/cursorfont.h>
#include <X11/Xfuncs.h>
#include <X11/Xresource.h>

#include <Xm/Xm.h>
#include <Xm/Protocols.h>
#include <Xm/MwmUtil.h>
#include <Xm/Form.h>
#include <Xm/Frame.h>
#include <Xm/Label.h>
#include <Xm/MenuShell.h>
#include <Xm/CascadeB.h>
#include <Xm/DrawnB.h>
#include <Xm/TextF.h>
#include <Xm/RowColumn.h>
#include <Xm/List.h>
#include <Xm/Separator.h>
#include <Xm/MessageB.h>
#include <Xm/PushB.h>
#include <Xm/SelectioB.h>
#include <Xm/ToggleB.h>
#include <Xm/ScrolledW.h>
#include <Xm/PanedW.h>
#include <Xm/DragDrop.h>
#include <Xm/FileSB.h>

#define MAJOR_VERSION	1
#define MINOR_VERSION	3

#ifndef INADDR_NONE
#define INADDR_NONE     0xffffffff
#endif

#define BUFFER_LENGTH   1056
#define NAME_LENGTH     256
#define SL_WIDTH        200
#define INIT_HEIGHT	800
#define INIT_FTP_HEIGHT	500
#define OFFSET		27

#define LOCAL_FILESYSTEM	10
#define FTP_FILESYSTEM		11

#define ACC_DROP_SITE		0
#define DIR_DROP_SITE		1
#define DEV_DROP_SITE		2

#define TRANSFER_ASCII		0
#define TRANSFER_BINARY		1

#define IS_EXECUTABLE(x)   ( ( x ) & ( S_IXUSR | S_IXGRP | S_IXOTH ) )

#ifndef S_ISLNK
#define S_ISLNK(x)	( ( x & S_IFMT ) == S_IFLNK )
#endif


typedef struct DragInfo
{
  Pixmap                p;
  int                   numDragNames;
  char                  dirName[NAME_LENGTH], *dragName;
} DragInfo;

struct FileList
{
  char                  name[BUFFER_LENGTH];
  char			perm[25];
  struct stat           s;
  struct FileList*      next;
};

struct DirInfo
{
  Widget                list, form, label, nameLabel;
  DragInfo              di;
  struct FileList*      fileList;
  int                   nFiles;
  struct stat           s;
  struct DirInfo*	next;
};

struct FtpDataConn
{
  XtInputId		acceptId;
  int			dataSocketId, acceptSocketId;
  int			index;
  char			*cmd, *results, recvDir[1056];
  FILE			*fd;
  struct FtpDataConn*	next;
};

struct FileManager
{
  Widget 		parent, mainForm, addAccelButton, topSW, topSWRC,
  			menubar, panedWindow, topForm, bottomForm, 
			devForm, devSW, devSWRC, transferAsciiTB, 
			transferBinaryTB;
  Widget		sw, clipWindow, sb, form;
  Boolean		runInXterm, isManaged;
  int			type, nDirInfo, nTotal, ht, transferType;
  struct DirInfo*	dirInfo;
  char			host[MAXHOSTNAMELEN], user[50], pass[50], *cwd, *cwdPtr;
  int			cwdIndex;
  XtInputId		cmdId;
  int			cmdSocketId;
  struct FtpDataConn*	ftpDataConn;
  struct FileManager*	next;
};

struct Accelerator
{
  DragInfo              di;
  struct stat           s;
  struct Accelerator    *next;
};

struct Children
{
  pid_t pid;
  XtInputId id;
  char *cmd;
  struct Children *next;
};

struct Device
{
  char name[256], command[256], pixmapName[256];
  Pixmap p;
  struct Device *next;
};

struct Rule
{
  char name[256], type[25], command[256], pixmapName[256];
  Pixmap pixmap;
  struct Rule *next;
};


#ifndef NeedFunctionPrototypes
#if defined(FUNCPROTO) || defined(__STDC__) || defined(__cplusplus) || defined(c_plusplus)
#define NeedFunctionPrototypes 1
#else
#define NeedFunctionPrototypes 0
#endif 
#endif 


/*  defined in devices.c  */

void writeDeviceResources(
#if NeedFunctionPrototypes
	FILE*
#endif
);
void insertDev(
#if NeedFunctionPrototypes
	Widget, Widget, Widget
#endif
);
void deleteDevice(
#if NeedFunctionPrototypes
	int
#endif
);
void loadDevicesData(
#if NeedFunctionPrototypes
	Widget
#endif
);
void execDeviceCommand(
#if NeedFunctionPrototypes
	Widget, struct Device *, int, char*
#endif
);
void readDevApplicationResources(
#if NeedFunctionPrototypes
        Widget
#endif
);
void parseDevApplicationResources(
#if NeedFunctionPrototypes
        Widget, Pixel, Pixel
#endif
);
void createDeviceIcons(
#if NeedFunctionPrototypes
	Widget, Pixel, Pixel
#endif
);

/*  defined in info.c  */

void propertiesCB(
#if NeedFunctionPrototypes
        Widget, XtPointer, XtPointer
#endif
);

/*  defined in fileMgr.c  */

int getDirIndexByDirName(
#if NeedFunctionPrototypes
	struct FileManager*, char*
#endif
);
struct FileManager* getFileMgrByFtpHost( 
#if NeedFunctionPrototypes
	char*
#endif
);
Boolean isaDeviceWidget(
#if NeedFunctionPrototypes
        Widget
#endif
);
Boolean isAddAccWidget(
#if NeedFunctionPrototypes
        Widget
#endif
);
void updateAllFileManagers(
#if NeedFunctionPrototypes
	void
#endif
);
void allocateFileManager( 
#if NeedFunctionPrototypes
	Widget, int, char*, char*, char* 
#endif
);
void deleteFileManager(
#if NeedFunctionPrototypes
        Widget
#endif
);

/*  defined in accels.c  */

void addAccelerator(
#if NeedFunctionPrototypes
        DragInfo*
#endif
);
void createDelAccelPopup(
#if NeedFunctionPrototypes
        Widget, XtPointer, XtPointer
#endif
);
void deleteAcceleratorByName(
#if NeedFunctionPrototypes
        char*
#endif
);
void readRCFile(
#if NeedFunctionPrototypes
        Widget
#endif
);
void loadAccels(
#if NeedFunctionPrototypes
        Widget
#endif
);
void writeRCFile(
#if NeedFunctionPrototypes
        void
#endif
);

/*  defined in dd.c  */

void initDD(
#if NeedFunctionPrototypes
        Widget
#endif
);
void registerDragSite(
#if NeedFunctionPrototypes
        Widget
#endif
);
void unregisterDragSite(
#if NeedFunctionPrototypes
        Widget
#endif
);
void registerDropSite(
#if NeedFunctionPrototypes
        Widget, int
#endif
);

/*  defined in menubar.c  */

void createToolbar(
#if NeedFunctionPrototypes
        struct FileManager *, Widget
#endif
);
void createMenubar( 
#if NeedFunctionPrototypes
	struct FileManager *, int 
#endif
);
void quitCB(
#if NeedFunctionPrototypes
        Widget, XtPointer, XtPointer
#endif
);

/*  defined in fileSelectionBox.c  */

void updateFileManager(
#if NeedFunctionPrototypes
        struct FileManager*
#endif
);
void createFileSelectionBox( 
#if NeedFunctionPrototypes
        struct FileManager*, int 
#endif
);
void forkNexec(
#if NeedFunctionPrototypes
        Widget, struct stat *, char*, int
#endif
);
void chooseListItem(
#if NeedFunctionPrototypes
        Widget, int*, int, struct FileManager*
#endif
);
void accelSelect(
#if NeedFunctionPrototypes
	DragInfo*, struct FileManager*
#endif
);

/*  defined in util.c  */

char* createPausemeFile(
#if NeedFunctionPrototypes
	void
#endif
);
int issuePipeCommand( 
#if NeedFunctionPrototypes
	char* 
#endif
);
void createErrorAndWarningDialogs( 
#if NeedFunctionPrototypes
	Widget 
#endif
);
void popupErrorDialog( 
#if NeedFunctionPrototypes
	char * 
#endif
);
void popupMessageDialog( 
#if NeedFunctionPrototypes
	char * 
#endif
);
void setWatchCursor( 
#if NeedFunctionPrototypes
	void 
#endif
);
void unsetWatchCursor( 
#if NeedFunctionPrototypes
	void 
#endif
);
void setWatchCursorForFM(
#if NeedFunctionPrototypes
        struct FileManager*
#endif
);
void unsetWatchCursorForFM(
#if NeedFunctionPrototypes
        struct FileManager*
#endif
);
void traverseCB(
#if NeedFunctionPrototypes
        Widget, XtPointer, XtPointer
#endif
);

/*  defined in typesFile.c  */

void configCB(
#if NeedFunctionPrototypes
        Widget, XtPointer, XtPointer
#endif
);
void insertFileName(
#if NeedFunctionPrototypes
        char*, char*
#endif
);
Pixmap isaDevice( 
#if NeedFunctionPrototypes
	char*
#endif
);
void readApplicationResources(
#if NeedFunctionPrototypes
        Widget
#endif
);
void parseApplicationResources(
#if NeedFunctionPrototypes
        Widget, Pixel, Pixel
#endif
);
void getCommand(
#if NeedFunctionPrototypes
        char*, char*
#endif
);
int getDefaultCommand(
#if NeedFunctionPrototypes
        char*, char*
#endif
);
Pixmap getPixmap(
#if NeedFunctionPrototypes
        char*, struct stat, int, char*
#endif
);
void setToolBarPixmaps(
#if NeedFunctionPrototypes
        Widget, Widget, Widget, Widget, Widget, Widget, Widget
#endif
);

/*  defined in pty.c  */

void launchCmd(
#if NeedFunctionPrototypes
        Widget, char*
#endif
);
void popupCmdDialog(
#if NeedFunctionPrototypes
        Widget, XtPointer, XtPointer
#endif
);
void popupCmdWithCmd(
#if NeedFunctionPrototypes
        Widget, char*
#endif
);

/*  defined in choosers.c  */

Widget createColorChooser(
#if NeedFunctionPrototypes
	 Widget, void (*okCBptr)(Widget,XtPointer,XtPointer),
         void (*applyCBptr)(Widget,XtPointer,XtPointer),
         void (*cancelCBptr)(Widget,XtPointer,XtPointer)
#endif
);
void destroyColorChooser(
#if NeedFunctionPrototypes
	void
#endif
);
Widget createFontChooser(
#if NeedFunctionPrototypes
         Widget, void (*okCBptr)(Widget,XtPointer,XtPointer),
         void (*applyCBptr)(Widget,XtPointer,XtPointer),
         void (*cancelCBptr)(Widget,XtPointer,XtPointer)
#endif
);
void destroyFontChooser(
#if NeedFunctionPrototypes
	void
#endif
);

/*  defined in ftp.c  */

int ftpConnect(
#if NeedFunctionPrototypes
	struct FileManager*
#endif
);
void ftpSendCommand(
#if NeedFunctionPrototypes
        struct FileManager*, char*, int, char*
#endif
);
void establishFtpDataConn(
#if NeedFunctionPrototypes
	XtPointer, int*, XtInputId*
#endif
);
void readFtpData(
#if NeedFunctionPrototypes
        XtPointer, int*, XtInputId*
#endif
);
void readFtpMessages(
#if NeedFunctionPrototypes
        XtPointer, int*, XtInputId*
#endif
);


#include "GenLListDefines/fileManager.def"
#include "GenLList.hpp"
#include "GenLListDefines/undef.def"

#include "GenLListDefines/dirInfo.def"
#include "GenLList.hpp"
#include "GenLListDefines/undef.def"

#include "GenLListDefines/fileList.def"
#include "GenLList.hpp"
#include "GenLListDefines/undef.def"

#include "GenLListDefines/accelList.def"
#include "GenLList.hpp"
#include "GenLListDefines/undef.def"

#include "GenLListDefines/ftpDataConnList.def"
#include "GenLList.hpp"
#include "GenLListDefines/undef.def"

#include "GenLListDefines/childList.def"
#include "GenLList.hpp"
#include "GenLListDefines/undef.def"

#include "GenLListDefines/deviceList.def"
#include "GenLList.hpp"
#include "GenLListDefines/undef.def"

#include "GenLListDefines/ruleList.def"
#include "GenLList.hpp"
#include "GenLListDefines/undef.def"
